(* file: dfa.sml *)
(* description: deterministic finite automata *)
(* author: Takahito Aoto *)

signature DFA = 
sig 
    type state = string
    type alphabet = string list
    type transition = ((state * string) * state) list
    type dfa = state list * alphabet * transition * state * state list
    val toString: dfa -> string
    val rename: string * int -> dfa -> dfa
    val reach: transition -> state * string list -> state
    val accept: string list -> dfa -> bool
    val travel: string list -> dfa -> unit
    val complement: dfa -> dfa
    val union: dfa * dfa -> dfa
    val intersection: dfa * dfa -> dfa
    val empty: dfa -> bool
    val subset: dfa * dfa -> bool
    val equivalent: dfa * dfa -> bool
    val minimize: dfa -> dfa
end

structure Dfa : DFA =
struct
local 
    structure L = List
    structure LP = ListPair
    structure U = Util
in

type state = string
type alphabet = string list
type transition = ((state * string) * state) list
type dfa = state list * alphabet * transition * state * state list

fun alpha (Q,Sigma,delta,q0,F) = Sigma

fun toString (Q,Sigma,delta,q0,F) = 
    let fun prRule ((q,l),r) = "   (" ^ q ^ "," ^ l ^ ") -> " ^ r ^ "\n"
    in "(STATES " ^ (U.toStringSpace (fn x=>x) Q) ^ ")\n"
       ^ "(ALPHABET " ^ (U.toStringSpace (fn x=>x) Sigma) ^ ")\n"
       ^ "(TRANSITIONS\n"
       ^ (String.concat (L.map prRule delta))
       ^ ")\n"
       ^"(START " ^ q0 ^ ")\n"
       ^"(ACCEPT " ^ (U.toStringSpace (fn x=>x) F) ^ ")\n"
    end

(* rename states of automaton to prefix ^ num, prefix ^ (num+1), .... *)
fun rename (prefix,num) (Q,Sigma,delta,q0,F) =
    let val renamingTable = ListPair.zip (Q, List.tabulate (List.length Q, fn x=> prefix ^ Int.toString (num+x)))
	fun rename q = valOf (U.lookup q renamingTable)
	val Q' = List.map rename Q
	val delta' = List.map (fn ((q,x),r) => ((rename q, x), rename r)) delta
	val q0' = rename q0
	val F' = List.map rename F
    in (Q',Sigma,delta',q0',F')
    end


fun step delta (q,l) = case U.lookup (q,l) delta of
			   SOME q' => q'
			 | NONE => raise Fail "invalid input string or invalid transition function?"

fun reach delta (q, []) = q
  | reach delta (q, l::ls) = reach delta (step delta (q,l), ls)

fun accept ls (_,_,delta,q0,F) = U.member (reach delta (q0,ls)) F

fun travel ls (_,_,delta,q0,F) =
    let fun printStep [] q = print ("(" ^ q ^ "," ^ "<EPSILON>)\n")
	  | printStep (c::cs) q = let val _ = print ("(" ^ q ^ "," ^ (String.concat (c::cs)) ^ ")\n->")
				     in printStep cs (step delta (q,c))
				     end
    in if null ls
       then print ("(" ^ q0 ^ "," ^ "<EPSILON>)\n")
       else printStep ls q0
    end

fun complement (Q,Sigma,delta,q0,F) = (Q,Sigma,delta,q0,U.setminus (Q,F))

fun getSeparator Q = 
    let fun checkSep sep = if L.exists (fn str => String.isSubstring sep str) Q
			   then checkSep (":" ^ sep) 
			   else sep
    in checkSep ":" 
    end

fun pairingStates sep (q1,q2) = q1 ^ sep ^ q2

fun union ((Q,Sigma,delta,q0,F), (Q',Sigma',delta',q0',F')) = 
    if not (U.seteq (Sigma,Sigma')) then raise Fail "union of DFAs of different alphabets"
    else let val sep = getSeparator (Q @ Q')
	 in (
	     ListXProd.mapX (pairingStates sep) (Q,Q'),
	     Sigma,
	     U.mapAppend (fn ((q,c),r) => 
			     List.mapPartial (fn ((q',c'),r') => if c = c' then SOME ((pairingStates sep (q,q'),c), pairingStates sep (r,r'))
								 else NONE) delta')
			 delta,
	     pairingStates sep (q0,q0'),
	     ListXProd.mapX (pairingStates sep) (F,F') 
	     @ ListXProd.mapX (pairingStates sep) (U.setminus (Q,F),F')
	     @ ListXProd.mapX (pairingStates sep) (F, U.setminus (Q',F'))
	    )
	 end

fun intersection ((Q,Sigma,delta,q0,F),(Q',Sigma',delta',q0',F')) = 
    if not (U.seteq (Sigma,Sigma')) then raise Fail "intersection of DFAs of different alphabets"
    else let val sep = getSeparator (Q @ Q')
	 in (
	     ListXProd.mapX (pairingStates sep) (Q,Q'),
	     Sigma,
	     U.mapAppend (fn ((q,c),r) => 
			     List.mapPartial (fn ((q',c'),r') => if c = c' then SOME ((pairingStates sep (q,q'),c), pairingStates sep (r,r'))
								 else NONE) delta')
			 delta,
	     pairingStates sep (q0,q0'),
	     ListXProd.mapX (pairingStates sep) (F,F')
	    )
	 end

(* emptiness check of DFA *)
fun reachableStates delta qs = 
    let fun neiboughr q = List.mapPartial (fn ((q',_),r) => if q' = q then SOME r else NONE) delta
	val new = U.union (qs, U.mapAppend neiboughr qs)
    in if U.subseteq (new, qs) then new
       else reachableStates delta new
    end

fun empty (_,_,delta,q0,F) = U.disjoint (reachableStates delta [q0],  F)

(* subseteq check of DFA *)
(* A \subseteq B iff A \cap non-B = emptyset *)
fun subset (M1,M2) = 
    if not (U.seteq (alpha M1, alpha M2)) then raise Fail "subset comparison of DFAs of different alphabets"
    else empty (intersection (M1, complement M2))

(* equivalence check of DFA *)
fun equivalent (M1,M2) =
    if not (U.seteq (alpha M1,alpha M2)) then raise Fail "equivalent comparison of DFAs of different alphabets"
    else subset (M1,M2) andalso subset (M2,M1)

(* minimizing DFA states *)
fun getUnmarked delta Sigma (marked, all) =
    let fun isMarked (p,q) = L.exists (fn (x,y) => (p,q) = (x,y) orelse (p,q) = (y,x)) marked
	val new = L.filter (fn (p,q) => L.exists (fn l => isMarked (step delta (p,l), step delta (q,l))) Sigma) all
    in if U.subseteq (new,marked) 
       then U.setminus (all, marked) (* saturated *)
       else getUnmarked delta Sigma (U.union (marked,new), all)
    end

fun collapseStates (p1,p2) (Q,Sigma,delta,q0,F) = 
    let fun replace p = if p = p2 then p1 else p
    in (U.setminus (Q, [p2]), Sigma, 
	U.removeDuplication  (L.map (fn ((q,x),r) => ((replace q,x),replace r)) delta),
	replace q0, U.removeDuplication  (L.map replace F))
    end
						    
fun reflectUnmark [] M = M
  | reflectUnmark ((p,q)::rest) M =  reflectUnmark rest (collapseStates (p,q) M)

fun minimize (Q,Sigma,delta,q0,F) =
    let (* remove inaccessbile states *)
	val Q' = reachableStates delta [q0]
	val delta' = L.filter (fn ((q,x),r) => U.member q Q' andalso U.member r Q') delta
	val F' = U.intersection (F,Q')
	(* states collapsing *)
	val all = U.orderedPairs Q'
	fun check (p,q) = (U.member p F' andalso not (U.member q F'))
			  orelse (not (U.member p F') andalso U.member q F')
	val marked = L.filter check all
	val unmarked = getUnmarked delta' Sigma (marked,all)
	val minM = reflectUnmark unmarked (Q',Sigma,delta',q0,F')
    in minM
    end



end (* of local *)
end
