(* file: nfa.sml *)
(* description: deterministic finite automata *)
(* author: Takahito Aoto *)

signature NFA = 
sig 
    type state = string
    type alphabet = string list
    type transition = ((state * string) * state list) list
    type nfa = state list * alphabet * transition * state * state list
    val toString: nfa -> string
    val accept: string list -> nfa -> bool
    val travel: string list -> nfa -> unit
    val showTransitionToAccept: string list -> nfa -> unit
    val powerset: nfa -> Dfa.dfa (* subset construction *)
    val rename: string * int -> nfa -> nfa
    val union: nfa * nfa -> nfa
    val concat: nfa * nfa -> nfa
    val star: nfa -> nfa
    val collapse: nfa -> nfa (* taking epsilon closure *)
end

structure Nfa : NFA =
struct
local 
    structure L = List
    structure LP = ListPair
    structure U = Util
in

type state = string
type alphabet = string list
type transition = ((state * string) * state list) list
type nfa = state list * alphabet * transition * state * state list

fun alpha (Q,Sigma,rho,q0,F) = Sigma

fun toString (Q,Sigma,rho,q0,F) = 
    let fun prRule ((q,l),rs) = "   (" ^ q ^ "," ^ l ^ ") -> " ^ (U.toStringSpace (fn x=>x) rs) ^ "\n"
    in "(STATES " ^ (U.toStringSpace (fn x=>x) Q) ^ ")\n"
       ^ "(ALPHABET " ^ (U.toStringSpace (fn x=>x) Sigma) ^ ")\n"
       ^ "(TRANSITIONS\n"
       ^ (String.concat (L.map prRule rho))
       ^ ")\n"
       ^"(START " ^ q0 ^ ")\n"
       ^"(ACCEPT " ^ (U.toStringSpace (fn x=>x) F) ^ ")\n"
    end

fun trans rho (q,l) = case U.lookup (q,l) rho of
			 SOME qs' => qs'
		       | NONE => []

(* transition *)
(* use rho for NFA transition function *)
fun epsilonClosure rho qs = 
    let val new = U.union (qs, U.mapUnion (fn q => trans rho (q,"<EPSILON>")) qs)
    in if U.subseteq (new, qs)
       then new
       else epsilonClosure rho new
    end

(* qs should be epsilon closed *)
fun step rho (qs,c) = epsilonClosure rho (U.mapUnion (fn q => trans rho (q,c)) qs)

fun reach rho (qs,[]) = qs
  | reach rho (qs, c::cs) = reach rho (step rho (qs,c), cs)

fun accept ls (_,_,rho,q0,F) = let val reachables = reach rho (epsilonClosure rho [q0], ls)
			       in not (null (U.intersection (reachables, F)))
			       end

fun travel ls (_,_,rho,q0,F) =
    let val show = U.toStringCommaBrace (fn x=>x)
	fun printStep [] qs = print ("(" ^ (show qs)^ "," ^ "<EPSILON>)\n")
	  | printStep (c::cs) qs = let val _ = print ("(" ^ (show qs)^ "," ^ (String.concat (c::cs)) ^ ")\n->")
				   in printStep cs (step rho (qs,c))
				   end
	val qs0 = epsilonClosure rho [q0]
    in if null ls
       then print ("(" ^ (show qs0)^ "," ^ "<EPSILON>)\n")
       else printStep ls qs0
    end

fun first (x,y) = x

fun removeDupByHd [] = []
  | removeDupByHd (ps::qss) = if L.exists (fn qs => first (hd ps) = first (hd qs)) qss
			      then removeDupByHd qss
			      else ps :: removeDupByHd qss

fun epsilonClosureWithHistory rho qss  = 
    let val new = U.mapAppend (fn qs => L.map (fn p => (p,0)::qs) (trans rho (first (hd qs),"<EPSILON>"))) qss
    in if U.subseteq (L.map (first o hd) new, L.map (first o hd) qss)
       then qss
       else epsilonClosureWithHistory rho (removeDupByHd (new @ qss))
    end

fun stepWithHistory rho (qss,c) =
    let val next = removeDupByHd (U.mapAppend (fn qs => L.map (fn p => (p,1)::qs) (trans rho (first (hd qs),c))) qss)
    in epsilonClosureWithHistory rho next
    end

fun reachWithHistory rho (qss,[]) = qss
  | reachWithHistory rho (qss, c::cs) = reachWithHistory rho (stepWithHistory rho (qss,c), cs)

fun acceptWithHistory ls (_,_,rho,q0,F) =
    let val reachables = reachWithHistory rho (epsilonClosureWithHistory rho [[(q0,0)]], ls)
    in L.find (fn qs => U.member (first (hd qs)) F) reachables
    end

fun showTransitionToAccept ls N =
    case acceptWithHistory ls N of
	NONE => raise Fail "request of accepted transition for rejected input"
      | SOME qs => case rev qs of
		       [] => raise Fail "request of accept transition for not accepted string"
		     | ((q0,_)::steps) => let fun show cs = if null cs then "<EPSILON>" else String.concat cs
					     val _ = print ("(" ^ q0 ^ "," ^ (show ls) ^ ")\n")
					     fun printStep [] [] = ()
					       | printStep _  [] = raise Fail "yet unreaded input remains"
					       | printStep cs ((q,i)::qs) = 
						 let val cs2 = if i = 0 then cs else tl cs
						     val _ = print ("->(" ^ q ^ "," ^ (show cs2)^ ")\n")
						 in printStep cs2 qs
						 end
					 in printStep ls steps
					 end

(* powerset construction *)
fun getSeparator Q = 
    let fun checkSep sep = if L.exists (fn str => String.isSubstring sep str) Q
			   then checkSep (":" ^ sep) 
			   else sep
    in checkSep ":" 
    end

fun powerset (Q,Sigma,rho,q0,F) =
    let val sep = getSeparator Q
	fun mkDfaState qs = ("{" ^ (String.concatWith sep (ListMergeSort.sort String.> qs)) ^ "}")
	val powerOfQ = U.powerset Q
	val Q' = List.map mkDfaState powerOfQ
	val delta' = ListXProd.mapX 
			 (fn (qs,c) => let val next = epsilonClosure rho (U.mapUnion (fn q => trans rho (q,c)) qs)
				       in ((mkDfaState qs, c), mkDfaState next)
				       end)
			 (powerOfQ, Sigma)
	val q0' = mkDfaState (epsilonClosure rho [q0])
	val F' = List.mapPartial (fn qs=> if not (null (U.intersection (qs,F))) then SOME (mkDfaState qs) else NONE) powerOfQ
    in (Q',Sigma,delta',q0',F')
    end

(* taking epsilon closure *)
fun collapse (Q,Sigma,rho,q0,F) =
    let val sep = getSeparator Q
	fun mkState qs = ("{" ^ (String.concatWith sep (ListMergeSort.sort String.> qs)) ^ "}")
	val qs0 = epsilonClosure rho [q0]
	fun addTrans0 (qs,x) = let val ps = U.mapUnion (fn ((q,y),rs) => if U.member q qs andalso x = y then rs else []) rho
				   val Ps = L.map (fn q => epsilonClosure rho [q]) ps
			       in if null Ps then NONE else SOME ((qs,x),Ps) 
			       end
	fun addTrans qs = L.mapPartial (fn x => addTrans0 (qs,x)) Sigma
	fun addStates (Qs, Rho, []) = (Qs,Rho)
	  | addStates (Qs, Rho, qs::rest) = 
	    if L.exists (fn rs => U.seteq (rs,qs)) Qs
	    then addStates (Qs, Rho, rest)
	    else let val Rho' = addTrans qs 
		     val Qs' = U.mapAppend (fn ((_,_),Ps) => Ps) Rho'
		 in addStates (qs::Qs, Rho@Rho', rest @ Qs')
		 end
	val (Qs,Rho) = addStates ([], [], [qs0])
	val Fs  = L.filter (fn qs => not (null (U.intersection (qs,F)))) Qs
	val Q'  = L.map mkState Qs
	val rho' = L.map (fn ((qs,x),Ps) => ((mkState qs, x), L.map mkState Ps)) Rho
	val q0' = mkState qs0
	val F' = L.map mkState Fs
    in (Q',Sigma,rho',q0',F')
    end

(* rename states of automaton to prefix ^ num, prefix ^ (num+1), .... *)
fun rename (prefix,num) (Q,Sigma,rho,q0,F) =
    let val renamingTable = ListPair.zip (Q, List.tabulate (List.length Q, fn x=> prefix ^ Int.toString (num+x)))
	fun rename q = valOf (U.lookup q renamingTable)
	val Q' = List.map rename Q
	val rho' = ListMergeSort.sort (fn (((q1,_),_),((q2,_),_)) => String.> (q1,q2))
				      (List.map (fn ((q,x),qs) => ((rename q, x), List.map rename qs)) rho)
	val q0' = rename q0
	val F' = List.map rename F
    in (Q',Sigma,rho',q0',F')
    end

fun union (N1, N2) = 
    if not (U.seteq (alpha N1, alpha N2)) then raise Fail "union of NFAs of different alphabets"
    else let val (Q1,Sigma,rho1,q1,F1) = rename ("q",1) N1
	     val num = List.length Q1
	     val q0 = "s"
	     val (Q2,_,rho2,q2,F2) = rename ("r",1) N2
	     val rho = [ ((q0,"<EPSILON>"),[q1,q2]) ]
			 @ rho1 @ rho2
	 in (q0::Q1@Q2, Sigma, rho, q0, F1 @ F2)
	 end

fun concat (N1,N2) = 
    if not (U.seteq (alpha N1, alpha N2)) then raise Fail "concatenation of NFAs of different alphabets"
    else let val (Q1,Sigma,rho1,q1,F1) = rename ("q",1) N1
	     val (Q2,_,rho2,q2,F2) = rename ("r",1) N2
	     val (change, stay) = L.partition (fn ((q,x),qs) => U.member q F1 andalso x = "<EPSILON>") rho1
	     val revised = L.map (fn q => case L.find (fn ((r,_),_) => q = r) change of
					      NONE => ((q,"<EPSILON>"),[q2])
					    | SOME ((q,x),qs) => ((q,x), U.union (qs,[q2])))
				 F1
	 in (Q1@Q2, Sigma, stay @ revised @ rho2, q1, F2)
	 end

fun star N =
    let val (Q,Sigma,rho,q1,F) = rename ("q",1) N
	val q0 = "q0"
	val (change, stay) = L.partition (fn ((q,x),qs) => U.member q F andalso x = "<EPSILON>") rho
	val revised = L.map (fn q => case L.find (fn ((r,_),_) => q = r) change of
					 NONE => ((q,"<EPSILON>"),[q1])
				       | SOME ((q,x),qs) => ((q,x), U.union (qs,[q1])))
			    F
	val new = [((q0,"<EPSILON>"),[q1])]
    in (q0::Q, Sigma, stay @ new @ revised, q0, q0::F)
    end


end (* of local *)
end
