(* file: reg.sml *)
(* description: regular expression *)
(* author: Takahito Aoto *)

signature REG = 
sig 
    datatype expr = 
	     SYMBOL of string 
	     | ANY
	     | CONCAT of expr list
	     | UNION of expr list
	     | STAR of expr
	     | ITER of expr
    val simplify : expr -> expr
    val toString : expr -> string
    val expr2nfa : string list -> expr -> Nfa.nfa
    val symbols: expr -> string list
    val dfa2re: bool -> Dfa.dfa -> string
end

structure Reg : REG =
struct
local 
    structure L = List
    structure LP = ListPair
    structure U = Util
    structure S = String
in

datatype expr = 
	 SYMBOL of string 
	 | ANY
	 | CONCAT of expr list
	 | UNION of expr list
	 | STAR of expr
	 | ITER of expr

fun isEmpty (SYMBOL "<EMPTY>") = true
  | isEmpty _ = false

fun isEpsilon (SYMBOL "<EPSILON>") = true
  | isEpsilon _ = false

fun simplify (UNION xs) = let val ys = L.filter (not o isEmpty) (L.map simplify xs)
			      val ys2 = U.mapAppend (fn y => case y of UNION ys => ys | _ => [y]) ys
			  in case ys2 of
				      [] => SYMBOL "<EMPTY>"
				    | [x] => x
				    | xs => UNION xs
			  end
  | simplify (CONCAT xs) = let val ys = L.filter (not o isEpsilon) (L.map simplify xs)
			       val ys2 = U.mapAppend (fn y => case y of CONCAT ys => ys | _ => [y]) ys
			   in if L.exists isEmpty ys2
			      then SYMBOL "<EMPTY>"
			      else case ys2 of
				       [] => SYMBOL "<EPSILON>"
				     | [x] => x
				     | xs => CONCAT xs
			   end
  | simplify (STAR x) = (case simplify x of
			    STAR y => STAR y
			  | ITER y => ITER y
			  | SYMBOL "<EPSILON>" => SYMBOL "<EPSILON>"
			  | SYMBOL "<EMPTY>" => SYMBOL "<EPSILON>"
			  | y => STAR y)
  | simplify (ITER x) = (case simplify x of
			    STAR y => STAR y
			  | ITER y => ITER y
			  | SYMBOL "<EPSILON>" => SYMBOL "<EPSILON>"
			  | SYMBOL "<EMPTY>" => SYMBOL "<EMPTY>"
			  | y => ITER y)
  | simplify ANY = ANY
  | simplify (SYMBOL str) = SYMBOL str

fun toString expr = 
    let fun toStringMain (UNION []) = "<EMPTY>"
	  | toStringMain (UNION [x]) = toStringMain x
	  | toStringMain (UNION xs) = S.concatWith "|" (L.map toStringMain xs)
	  | toStringMain (CONCAT []) = "<EPSILON>"
	  | toStringMain (CONCAT [x]) = toStringMain x
	  | toStringMain (CONCAT xs) = let fun toStringSub (UNION xs) = "(" ^ (toStringMain (UNION xs)) ^ ")"
					     | toStringSub x = toStringMain x
				       in S.concat (L.map toStringSub xs)
				       end
	  | toStringMain (STAR x) = let fun toStringSub (UNION xs) = "(" ^ (toStringMain (UNION xs)) ^ ")"
					  | toStringSub (CONCAT xs) = "(" ^ (toStringMain (CONCAT xs)) ^ ")"
					  | toStringSub x = toString x
				    in (toStringSub x) ^ "*"
				    end
	  | toStringMain (ITER x) = let fun toStringSub (UNION xs) = "(" ^ (toStringMain (UNION xs)) ^ ")"
					  | toStringSub (CONCAT xs) = "(" ^ (toStringMain (CONCAT xs)) ^ ")"
					  | toStringSub x = toString x
				    in (toStringSub x) ^ "+"
				    end
	  | toStringMain ANY = "."
	  | toStringMain (SYMBOL str) = str
    in toStringMain (simplify expr)
    end

fun symbols (UNION xs) = U.mapUnion symbols xs
  | symbols (CONCAT xs) = U.mapUnion symbols xs
  | symbols (STAR x) = symbols x
  | symbols (ITER x) = symbols x
  | symbols ANY = []
  | symbols (SYMBOL str) = [str]

fun expr2nfa Sigma expr =
    let val simp = simplify expr
	val count = ref 0
	fun newState () = (count := (!count) + 1; "q" ^ (Int.toString (!count)))
	fun makeNfa ANY = let val p = newState ()
			      val q = newState ()
			  in ([p,q],Sigma, L.map (fn x => ((p,x),[q])) Sigma, p, [q])
			  end
	  | makeNfa (SYMBOL str) = if str = "<EPSILON>"
				   then let val p = newState ()
					in ([p], Sigma, [], p, [p])
					end
				   else if str = "<EMPTY>"
				   then let val p = newState ()
					in ([p], Sigma, [], p, [])
					end
				   else let val p = newState ()
	  				    val q = newState ()
	  				in ([p,q],Sigma, [((p,str),[q])], p, [q])
	  				end
	  | makeNfa (CONCAT es) = (case es of
	  			      [] => raise Fail "NFA not yet simplified?\n"
	  			    | [e] => makeNfa e
	  			    | (e1::es2) => let val N1 = makeNfa e1
	  					       val N2 = makeNfa (CONCAT es2)
	  					   in Nfa.concat (N1,N2)
	  					   end)
	  | makeNfa (UNION es) = (case es of
	  			      [] => raise Fail "NFA not yet simplified?\n"
	  			    | [e] => makeNfa e
	  			    | (e1::es2) => let val N1 = makeNfa e1
	  					       val N2 = makeNfa (UNION es2)
	  					   in Nfa.union (N1,N2)
	  					   end)
	  | makeNfa (STAR e) = let val N = makeNfa e
	  		       in Nfa.star N
	  		       end
	  | makeNfa (ITER e) = let val N1 = makeNfa e
	  			   val N2 = Nfa.star N1
	  		       in Nfa.concat (N1,N2)
	  		       end
    in makeNfa simp
    end


fun gnfa2str (Q,Sigma,delta,q0,q1) = 
    let fun prRule ((p,r),s) = "   (" ^ p ^ "," ^ s ^ ") -> "  ^ r ^ "\n"
    in "(STATES " ^ (U.toStringSpace (fn x=>x) Q) ^ ")\n"
       ^ "(ALPHABET " ^ (U.toStringSpace (fn x=>x) Sigma) ^ ")\n"
       ^ "(TRANSITIONS\n"
       ^ (String.concat (L.map prRule delta))
       ^ ")\n"
       ^"(START " ^ q0 ^ ")\n"
       ^"(ACCEPT " ^ q1 ^ ")\n"
    end

fun dfa2gnfa verbose M =
    let val (Q,Sigma,delta,q2,F) = Dfa.rename ("q",1) M
	val q0 = "q_start"
	val q1 = "q_accept"
	val delta1 = L.map (fn ((p,s),r) => ((p,r),s)) delta
	fun join [] = []
	  | join (x::xs) = joinSub x ([], xs)
	and joinSub ((p,q),s) (others,[]) = ((p,q),s) :: (join (rev others))
	  | joinSub ((p,q),s) (others,(h as ((p1,q1),s1))::rest) = 
	    if (p,q) = (p1,q1)
	    then joinSub ((p,q), (s ^ "|" ^ s1)) (others,rest)
	    else joinSub ((p,q),s) (h::others, rest)
	val delta' = ((q0,q2),"<EPSILON>") :: ((L.map (fn q => ((q,q1),"<EPSILON>")) F) @ (join delta1))
	val _ = if verbose 
		then (print "=== initial GNFA ===\n"; 
		      print (gnfa2str (q0::q1::Q, Sigma, delta', q0, q1)))
		else ()
    in (q0::q1::Q, Sigma, delta', q0, q1)
    end

fun reviseDelta delta qrip (qi,qj) = 
    let val R1 = case U.lookup (qi,qrip) delta of NONE => "<EMPTY>" | SOME str => str
	val R2 = case U.lookup (qrip,qrip) delta of NONE => "<EMPTY>" | SOME str => str
	val R3 = case U.lookup (qrip,qj) delta of NONE => "<EMPTY>" | SOME str => str
	val R4 = case U.lookup (qi,qj) delta of NONE => "<EMPTY>" | SOME str => str
	val R = "(" ^ R1 ^ ")" ^ "(" ^ R2 ^ ")*" ^ "(" ^ R3 ^ ")|"  ^ "(" ^ R4 ^ ")"
    in ((qi,qj),R)
    end

fun convert verbose (Q,Sigma,delta,start,accept) =
    if length Q = 2
    then let val ((q0,q1),str) = hd delta
	 in if (q0 <> start) orelse (q1 <> accept) 
	    then raise Fail "conversion of GNFA failed?\n"
	    else str
	 end
    else case L.find (fn q => q <> start andalso q <> accept) Q of
	     NONE => raise Fail "no state for rip?\n"
	   | SOME qrip => let val _ = if verbose  
				      then print ("=== ripping state " ^ qrip ^ " ===\n")
				      else ()
			      val Q' = U.setminus (Q, [qrip])
			      val delta' = ListXProd.mapX (reviseDelta delta qrip) (U.setminus (Q',[accept]), U.setminus (Q',[start]))
			      val _ = if verbose  
				      then print (gnfa2str (Q',Sigma,delta',start,accept))
				      else ()
			  in convert verbose (Q',Sigma,delta',start,accept)
			  end

fun dfa2re verbose M = convert verbose (dfa2gnfa verbose M)


end (* of local *)
end
